#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#define LTX_IGNORE_MISSING_INCLUDES



#include "ltx.inc"



/*
 * Raw chunk of .ltx data as table of text lines
 */
typedef struct
{
    char *name;
    int num_lines;
    int max_lines;
    char **lines;
} LTX_CHUNK;



static LTX_CHUNK* create_chunk(char *name)
{
    LTX_CHUNK *chunk_p = malloc(sizeof(LTX_CHUNK));

    memset(chunk_p, 0, sizeof(LTX_CHUNK));

    chunk_p->name = strdup(name);
    chunk_p->max_lines = 40;
    chunk_p->lines = malloc(chunk_p->max_lines * sizeof(char*));

    return chunk_p;
}


static void add_line_to_chunk(LTX_CHUNK *chunk_p, char *str_p)
{
    if (chunk_p->num_lines == chunk_p->max_lines)
    {
        chunk_p->max_lines += 40;
        chunk_p->lines = realloc(chunk_p->lines, chunk_p->max_lines * sizeof(char*));
    }

    chunk_p->lines[chunk_p->num_lines] = malloc(strlen(str_p) + 100 + 1);
    strcpy(chunk_p->lines[chunk_p->num_lines], str_p);
    chunk_p->num_lines++;
}


static void destroy_chunk(LTX_CHUNK *chunk_p)
{
    int i;

    free(chunk_p->name);

    for (i = 0; i < chunk_p->num_lines; i++)
        free(chunk_p->lines[i]);

    free(chunk_p->lines);
    free(chunk_p);
}


static char* ptr2_val4_key(char *str_p, const char *key)
{
    char *end_p = (char*)ltx_find_eol(str_p);

    for (; isspace(*str_p); str_p++);

    if (str_p >= end_p)
        return NULL;

    if (strstr(str_p, key) != str_p)
        return NULL;


    str_p = strchr(str_p, '=');

    if (str_p == NULL)
        return NULL;

    return str_p + 1;
}


static void write_chunk(LTX_CHUNK *chunk_p, FILE *fpout)
{
    int i;

    for (i = 0; i < chunk_p->num_lines; i++)
        fputs(chunk_p->lines[i], fpout);
}


/* Write config chunk to file unless it's one of the auto generated ones */
static void process_chunk(LTX_CHUNK *chunk_p, LTX_DATA *ltx_p, FILE *fpout, int *modified)
{
    LTX_SECT *sect_p = ltx_find_sect(ltx_p, chunk_p->name);
    LTX_LINE *line_p;
    char suffix[40];
    int i;

    /* Weight mod data */
    static const char *weight_suffix = NULL;
    static int weight_num_sect;
    static float weight_increment;
    static float weight2_offset;

    /* Weapon mod data */
    static const char *wpn_suffix = NULL;
    static int wpn_num_sect;
    static float dispersion_factor;


    if (sect_p == NULL)
    {
        printf("WARNING: can't find section [%s]\n", chunk_p->name);
        write_chunk(chunk_p, fpout);
        return;
    }


    /* First time load settings for weight mod */
    if (weight_suffix == NULL)
    {
        sect_p = ltx_find_sect(ltx_p, "weightmod");
    
        if (sect_p == NULL)
        {
            puts("ERROR: can't find setting config section");
            return;
        }

        line_p = ltx_find_line(sect_p, "suffix_format");
        if (line_p == NULL || line_p->val == NULL)
        {
            puts("ERROR: can't find suffix_format");
            return;
        }
        weight_suffix = line_p->val;

        line_p = ltx_find_line(sect_p, "num_sections");
        if (line_p == NULL || line_p->val == NULL)
        {
            puts("ERROR: can't find num_sections");
            return;
        }
        sscanf(line_p->val, "%d", &weight_num_sect);

        line_p = ltx_find_line(sect_p, "weight_increment");
        if (line_p == NULL || line_p->val == NULL)
        {
            puts("ERROR: can't find weight_increment");
            return;
        }
        sscanf(line_p->val, "%f", &weight_increment);
    }

    /* ... and for weapon mod */
    if (wpn_suffix == NULL)
    {
        sect_p = ltx_find_sect(ltx_p, "wpnmod");
    
        if (sect_p == NULL)
        {
            puts("ERROR: can't find setting config section");
            return;
        }

        line_p = ltx_find_line(sect_p, "suffix_format");
        if (line_p == NULL || line_p->val == NULL)
        {
            puts("ERROR: can't find suffix_format");
            return;
        }
        wpn_suffix = line_p->val;

        line_p = ltx_find_line(sect_p, "num_sections");
        if (line_p == NULL || line_p->val == NULL)
        {
            puts("ERROR: can't find num_sections");
            return;
        }
        sscanf(line_p->val, "%d", &wpn_num_sect);

        line_p = ltx_find_line(sect_p, "dispersion_factor");
        if (line_p == NULL || line_p->val == NULL)
        {
            puts("ERROR: can't find dispersion_factor");
            return;
        }
        sscanf(line_p->val, "%f", &dispersion_factor);

        /*printf("dispersion_factor = %f\n", dispersion_factor);*/
    }

    /* If the section name is one of the generated sections' name */

    for (i = 0; i < weight_num_sect; i++)
    {
        sprintf(suffix, weight_suffix, i);

        if (strstr(sect_p->name, suffix) != NULL)
            return;
    }


    for (i = 0; i < wpn_num_sect; i++)
    {
        sprintf(suffix, wpn_suffix, i);

        if (strstr(sect_p->name, suffix) != NULL)
            return;
    }


    write_chunk(chunk_p, fpout);

    
    /* Is this an outfit */
    if ((line_p = ltx_find_line(sect_p, "class")) != NULL
        && line_p->val != NULL
        && (strcmp(line_p->val, "E_STLK") == 0 || strcmp(line_p->val, "EQU_MLTR") == 0))
    {
        float add_weight  = 0.0;
        float add_weight2 = 0.0;

        line_p = ltx_find_line(sect_p, "additional_inventory_weight");

        if (line_p == NULL || line_p->val == NULL)
            return;

        sscanf(line_p->val, "%f", &add_weight);

        line_p = ltx_find_line(sect_p, "additional_inventory_weight2");

        if (line_p == NULL || line_p->val == NULL)
            return;

        sscanf(line_p->val, "%f", &add_weight2);


        for (i = 0; i <= weight_num_sect; i++)
        {
            sprintf(suffix, weight_suffix, i);
          
            fprintf(fpout,
                    "[%s%s]:%s\n"
                    "additional_inventory_weight  = %.1f\n"
                    "additional_inventory_weight2 = %.1f\n"
                    "quest_item                   = true\n"
                    "\n",
                    sect_p->name,
                    suffix,
                    sect_p->name,
                    add_weight,
                    add_weight2);

            add_weight  += weight_increment;
            add_weight2 += weight_increment;
        }

        *modified = 1;
    }
    else if ((line_p = ltx_find_line(sect_p, "ammo_mag_size")) != NULL && line_p->val != NULL)
    {
        float base_disp;
        float disp_decr;

        if (line_p->val == NULL || atoi(line_p->val) == 0)
            return;

        line_p = ltx_find_line(sect_p, "fire_dispersion_base");

        if (line_p == NULL || line_p->val == NULL)
            return;

        sscanf(line_p->val, "%f", &base_disp);

        disp_decr = dispersion_factor * base_disp / wpn_num_sect;

        /*printf("disp_decr = %.5f\n", disp_decr);*/


        for (i = 0; i <= wpn_num_sect; i++)
        {
            sprintf(suffix, wpn_suffix, i);
          
            fprintf(fpout,
                    "[%s%s]:%s\n"
                    "fire_dispersion_base = %.5f\n"
                    "quest_item           = true\n"
                    "\n",
                    sect_p->name,
                    suffix,
                    sect_p->name,
                    base_disp);

            base_disp -= disp_decr;
        }

        *modified = 1;
    }
}


static int modify_file(char *ltx_name, LTX_DATA *ltx_p)
{
    FILE *fpin;
    FILE *fpout;
    char bak_name[_MAX_PATH];
    LTX_CHUNK *chunk_p = NULL;

    int file_modified = 0;

    sprintf(bak_name, "%s.bak", ltx_name);

    if (rename(ltx_name, bak_name) != 0)
    {
        printf("Error renaming %s to %s", ltx_name, bak_name);
        perror("");
        return EXIT_FAILURE;
    }

    if ((fpin = fopen(bak_name, "r")) == NULL)
    {
        printf("Error opening file %s", bak_name);
        perror("");
        return EXIT_FAILURE;
    }

    if ((fpout = fopen(ltx_name, "w")) == NULL)
    {
        printf("Error opening file %s", ltx_name);
        perror("");
        return EXIT_FAILURE;
    }


    while (!feof(fpin))
    {
        char buf[MAX_FILE_LINE_LEN];
        char chunk_name[MAX_SECT_NAME_LEN];
        char *str_p = buf;

        if (fgets(buf, sizeof(buf), fpin) == NULL)
            break;

        for (; isspace(*str_p); str_p++);

        if (*str_p != 0 && sscanf(str_p, "[%[^]]s]", chunk_name) == 1)
        {
            if (chunk_p != NULL)
            {
                process_chunk(chunk_p, ltx_p, fpout, &file_modified);
                destroy_chunk(chunk_p);
            }
            chunk_p = create_chunk(chunk_name);
        }

        if (chunk_p != NULL)
        {
            add_line_to_chunk(chunk_p, buf);

            if (feof(fpin))
            {
                process_chunk(chunk_p, ltx_p, fpout, &file_modified);
                destroy_chunk(chunk_p);
                chunk_p = NULL;
            }
        }
        else
            fputs(buf, fpout);
    }

    if (chunk_p != NULL)
    {
        process_chunk(chunk_p, ltx_p, fpout, &file_modified);
        destroy_chunk(chunk_p);
    }

    fclose(fpin);
    fclose(fpout);


    if (!file_modified)
    {
        if (remove(ltx_name) != 0)
        {
            printf("Error removing %s", ltx_name);
            perror("");
            return EXIT_FAILURE;
        }
        

        if (rename(bak_name, ltx_name) != 0)
        {
            printf("Error renaming %s to %s", bak_name, ltx_name);
            perror("");
            return EXIT_FAILURE;
        }

        printf("%s unchanged\n", ltx_name);
    }

    return EXIT_SUCCESS;
}


int main(int argc, char** argv)
{
    int i;
    LTX_DATA ltx;
    LTX_SECT *sect_p;
    LTX_LINE *line_p;
    char path[_MAX_PATH];
    char *ch_p;

    /* Setup file name for program's config file */
    strcpy(path, argv[0]);

    ch_p = strrchr(path, '.');

    if (ch_p == NULL)
       strcat(path, ".ltx");
    else
       strcpy(ch_p, ".ltx");



    ltx_init_data(&ltx);


    /* Load own config */    
    if (ltx_load_file(&ltx, path) != LTX_SUCCESS)
    {
        printf("Error loading %s\n", path);
        return EXIT_FAILURE;
    }
    
    sect_p = ltx_find_sect(&ltx, "global");
    if (sect_p == NULL)
    {
        puts("Cannot find section [global]");
        return EXIT_FAILURE;
    }

    line_p = ltx_find_line(sect_p, "gamedata");
    if (line_p != NULL && line_p->val != NULL)
    {
        chdir(line_p->val);
    }

    sect_p = ltx_find_sect(&ltx, "load_files");

    if (sect_p != NULL)
    {
        for (i = 0; i < sect_p->num_lines; i++)
        {
            if (ltx_load_file(&ltx, sect_p->line_array[i]->val) != LTX_SUCCESS)
            {
                printf("Error loading %s\n", sect_p->line_array[i]->val);
                return EXIT_FAILURE;
            }
        }
    }


    ltx_complete_data(&ltx);


    sect_p = ltx_find_sect(&ltx, "modify_files");

    if (sect_p != NULL)
    {
        for (i = 0; i < sect_p->num_lines; i++)
        {
            if (modify_file(sect_p->line_array[i]->val, &ltx) != EXIT_SUCCESS)
            {
                printf("Error modifying %s\n", sect_p->line_array[i]->val);
                return EXIT_FAILURE;
            }
        }
    }

    ltx_clear_data(&ltx);

    return EXIT_SUCCESS;
}
